// TLM_GPU class implementionfor Cuda TLM project

/*
    Copyright 2010 Computational Electromagnetics Research Laboratory (CERL)
    University of Victoria, Victoria, B.C. Canada
    http://www.cerl.ece.uvic.ca
    http://www.cerl.ece.uvic.ca/Open_Source_CEM.htm
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "TLM_GPU.h"
#include "TLMSim.h"
#include <math.h>


// include standard header files
#include <iostream>
#include <cstdlib>


namespace GPU_Space{

extern "C" bool cuda_Device_Init_wID(int deviceID );
extern "C" int cuda_NumOfDevices_get();
extern "C" float* cuda_GPU_mesh_Mem_allocate(int meshDim_X,int meshDim_Y,int meshDim_Z);
extern "C" void cuda_mesh_CPU_to_GPU_copy_relay(float* d_GPU_Device_Mesh, float* h_mesh_data, int meshDim_X,int meshDim_Y,int meshDim_Z);
				
extern "C" void cuda_mesh_CPU_to_GPU_get(float* d_GPU_Device_Mesh, float* h_mesh_data, int meshDim_X,int meshDim_Y,int meshDim_Z);
extern "C" void cuda_Free_GPU_Mesh(float* GPU_Device_Mesh);
extern "C" void cuda_reflect_coef_constants_set_relay(float* reflection_coef_list, int boundaryNum);
extern "C" void cuda_Mesh_Block_constants_set_relay(int meshDimX,int meshDimY ,int meshDimZ ,int thread_blk_X ,int thread_blk_Y ,int thread_blk_Z );
extern "C" void cuda_iteration_relay(int iteration, float* GPU_Device_Mesh, iterationSetupType* iterSetup );
extern "C" void cuda_TLM_3D_InterchangeOnly(int iteration, float* GPU_Device_Mesh, iterationSetupType* iterSetup );
extern "C" void cuda_excitation_execute_relay(float* GPU_Device_Mesh, int scn_arm, int x_min,int x_max,int y_min,int y_max,int z_min,int z_max, float excitation, int iteration_index);
extern "C" void cuda_excitation_execute_Impulse_relay(float* GPU_Device_Mesh, int x,int y,int z, float excitation, int iteration_index);

extern "C" void cuda_sampling_execute_relay(float* GPU_Device_Mesh, float* GPU_Samples_ptr, int sample_index, int polerize_dir, int x, int y,int z, int iteration_index);
extern "C" void cuda_gpu_sampleArray_Allocate_relay(float** d_samples, int sample_numMax);
extern "C" void cuda_Samples_get_relay(float* d_samples,  float* h_samples, int num_samples_to_get);
extern "C" void cuda_save_results_to_CSVfile(char* filename, float samples_CUDA[], int sample_len, float delta_time);
extern "C" void cuda_Device_to_host_memcopy_relay(float* d_data, float*  h_data, int num_nodes_to_get);
extern "C" void cuda_Test();
extern "C" void cuda_boundary_execute_relay(float* GPU_Device_Mesh, int x_min,int x_max,int y_min,int y_max,int z_min,int z_max, float reflect_coef);

	//////////////////////////
	// Class GPU Definitions//
	//////////////////////////

	// Constructor for GPU Class
	GPU::GPU(){
		meshNodeSize=0;
		iteration_index = 0;
		iterations_MaxNum = 0;
	
	}// end of GPU()
	 
	// read in structure definition file
	GPU::GPU(char* filename, float scale_mesh){
		iteration_index = 0;
		iterations_MaxNum =0;
		iteration_timer_lastTime = 0;
		iteration_previously_saved_idx = 0;
		MegaNodes_per_sec_running_avg_savd = 0;
		read_Structure_Definition_file(filename);
		
		strucuture_scaling(scale_mesh);

		GPU::delta_time = (resolution)/(VEL_LIGHT*2);
		GPU::GPU_BLOCK_X = 64;
		GPU::GPU_BLOCK_Y = 1;
		GPU::GPU_BLOCK_Z = 1;
		Mesh_GPU_create(meshDim_X,meshDim_Y,meshDim_Z);
		bool initSucess = cuda_Initalize(-1);
		GPU::GPU_Device_Mesh  = cuda_GPU_mesh_Mem_allocate(meshDim_X ,meshDim_Y ,meshDim_Z); 
		//bndry_seed_boundaries(); 
		cuda_reflect_coef_constants_set();
		cuda_Mesh_Block_constants_set();
		cuda_mesh_CPU_to_GPU_copy(); 
		cuda_iteration_setup();
		iteration_timer_display_ENABLE = false;
	}
	
	// Constructor for GPU Class: defines dimensions of Mesh
	GPU::GPU(int mesh_X, int mesh_Y, int mesh_Z, float resolution){
		iteration_index = 0;
		iterations_MaxNum =0;
		iteration_timer_lastTime = 0;
		iteration_previously_saved_idx = 0;
		MegaNodes_per_sec_running_avg_savd = 0;
		GPU::resolution = resolution;
		GPU::delta_time = (resolution)/(VEL_LIGHT*2);
		Mesh_GPU_create(mesh_X,mesh_Y,mesh_Z);
		// instantiate device 
		bool initSucess = cuda_Initalize(-1);
		GPU_Device_Mesh  = cuda_GPU_mesh_Mem_allocate(meshDim_X ,meshDim_Y ,meshDim_Z);

	} // end of GPU(...)
	
	// Constructor for GPU Class: defines dimensions of Mesh and passes in CPU_Mesh object to copy from
	GPU::GPU(float *mesh_data_parent, int mesh_X, int mesh_Y, int mesh_Z, float resolution){
		iteration_index = 0;
		iterations_MaxNum =0;
		iteration_timer_lastTime = 0;
		iteration_previously_saved_idx = 0;
		MegaNodes_per_sec_running_avg_savd = 0;
		GPU::resolution = resolution;
		GPU::delta_time = (resolution)/(VEL_LIGHT*2);
		mesh_client = mesh_data_parent;	
		Mesh_GPU_create(mesh_X,mesh_Y,mesh_Z);
		// seed the GPU Mesh with client data
		// instantiate device 
		cuda_Initalize(-1);
		GPU_Device_Mesh  = cuda_GPU_mesh_Mem_allocate(meshDim_X ,meshDim_Y ,meshDim_Z);
		Mesh_Client2GPU_set(mesh_client, mesh_X, mesh_Y, mesh_Z);
		//cuda_mesh_CPU_to_GPU_copy(GPU_Device_Mesh, mesh_host, meshDim_X,meshDim_Y,meshDim_Z);

	} // end of GPU(...)
	
	// Same as above but also defines which GPU device index to intialize
	GPU::GPU(float *mesh_data_parent, int mesh_X, int mesh_Y, int mesh_Z, float resolution, int device){
		iteration_index = 0;
		iterations_MaxNum =0;
		iteration_timer_lastTime = 0;
		iteration_previously_saved_idx = 0;
		MegaNodes_per_sec_running_avg_savd = 0;
		GPU::resolution = resolution;
		GPU::delta_time = (resolution)/(VEL_LIGHT*2);
		mesh_client = mesh_data_parent;
		Mesh_GPU_create(mesh_X,mesh_Y,mesh_Z);
		// seed the GPU Mesh with client data
		// instantiate device with specified index
		cuda_Initalize(device);
		GPU_Device_Mesh  = cuda_GPU_mesh_Mem_allocate(meshDim_X ,meshDim_Y ,meshDim_Z);
		Mesh_Client2GPU_set(mesh_client, mesh_X, mesh_Y, mesh_Z);
		//cuda_mesh_CPU_to_GPU_copy(GPU_Device_Mesh, mesh_host, meshDim_X,meshDim_Y,meshDim_Z);
		
	}// end of GPU(...)
	
	GPU::~GPU(){
		// TBD: Excitation objects must be destroyed
		cuda_Free_GPU_Mesh(GPU_Device_Mesh);
	}// end of destructor


	///////////////////
	// GPU Cuda Test Method
	void GPU::GPU_cuda_Test(){
		cuda_Test();
	}
	//
	////////////////




	/////////////////
	// CUDA CODE CALLS (relays calls to cuda code)
	bool GPU::cuda_Initalize(int deviceID){
		return cuda_Device_Init_wID(deviceID );
	}

	int GPU::cuda_NumOfDevices_get(){
		return cuda_NumOfDevices_get();
	}

	void GPU::cuda_Mesh_Allocate(float* mesh_host,int meshDim_X,int meshDim_Y,int meshDim_Z){
		GPU_Device_Mesh = cuda_GPU_mesh_Mem_allocate(meshDim_X,meshDim_Y,meshDim_Z);
	}
	void GPU::cuda_mesh_CPU_to_GPU_copy(){
		cuda_mesh_CPU_to_GPU_copy_relay(GPU_Device_Mesh, mesh_host, meshDim_X,meshDim_Y,meshDim_Z);
		
	}
	//
	/////////////////


	/////////////////
	// MESH METHODS
	//
	int GPU::address(int voltage_arm, int x, int y, int z){
		return meshDim_Z*meshDim_Y*meshDim_X*(voltage_arm - 1) + meshDim_Y*meshDim_X*z + meshDim_X*y + x;
	}

	// instantiate an Mesh Data object of the GPU_Mesh type
	void GPU::Mesh_GPU_create(int x, int y, int z){
		int num_of_X_blocks = (x/64);
		if (x%64 != 0 ) num_of_X_blocks++;
		GPU::meshDim_X = num_of_X_blocks*64;
		GPU::meshDim_Y = y;
		GPU::meshDim_Z = z;
		GPU::meshNodeSize = meshDim_X*meshDim_Y*meshDim_Z;
		//mesh_host = (Varm*)malloc(meshNodeSize*sizeof(Varm));
		mesh_host = (float*)malloc(12*meshNodeSize*sizeof(float));
		for (int idx_Z = 0 ; idx_Z< meshDim_Z ; idx_Z++) {
			for (int idx_Y = 0 ; idx_Y< meshDim_Y ; idx_Y++) {
				for (int idx_X = 0 ; idx_X< meshDim_X  ; idx_X++) {
					for (int idx_Arm = 1 ; idx_Arm <= 12  ; idx_Arm++) {
						mesh_host[address(idx_Arm,idx_X,idx_Y,idx_Z)] = 0; 
					} // end of idx_Arm for next
				} // end of idx_X for next
			} // end of idx_Y for next

		} // end of idx_Z for next
		
		
		
		//for (int idx_Z = 0 ; idx_Z< meshDim_Z ; idx_Z++) {
			//for (int idx_Arm = 1 ; idx_Arm <= 12  ; idx_Arm++) {
			//	mesh_host[address(idx_Arm,5,5,9)] = 1; 		
			//}
		//}




	} // end of method



	// Pass the data from CPU_Mesh and 
	//   populate the GPU_Mesh with it
	void GPU::Mesh_Client2GPU_set(float *mesh_data_client, int mesh_X, int mesh_Y, int mesh_Z){
		// assuming that incoming data structure is a 3D array where node index is: 
		//   node address = 12*(  mesh_Y*mesh_X*z + mesh_X*y + x)
		if (meshNodeSize >0){ // check if mesh_data was instantiated
			for (int z_idx = 0 ; z_idx < mesh_Z; z_idx++){
				for (int y_idx = 0 ; y_idx < mesh_Y; y_idx++){
					for(int x_idx = 0 ; x_idx < mesh_X; x_idx++){
						for (int arm_idx = 1 ; arm_idx <= 12  ; arm_idx++) {
							mesh_host[address(arm_idx,x_idx,y_idx,z_idx)] = mesh_data_client[12*(mesh_Y*mesh_X*z_idx + mesh_X*y_idx + x_idx)+  0];
						}						
					}
				}
			}
		} else {
			printf(" GPU mesh structure needs to be created first\n");
		}
	}
	///// GETs
	// Get Mesh Data from GPU  by populating the CPU data
	void GPU::Mesh_GPU2Client_get(float *mesh_data_client, int *gpu_mesh_X, int *gpu_mesh_Y, int *gpu_mesh_Z){
		// assuming that incoming data structure is a 3D array where node index is: 
		//   node address = 12*(  mesh_Y*mesh_X*z + mesh_X*y + x)
		if ((meshDim_X*meshDim_Y*meshDim_Z) != meshNodeSize){
			printf("Client Mesh size does not match GPU Mesh size\n");
			return; 
		}
		if (meshNodeSize >0){ // check if mesh_data was instantiated
			for (int z_idx = 0 ; z_idx < meshDim_Z; z_idx++){
				for (int y_idx = 0 ; y_idx < meshDim_Y; y_idx++){
					for(int x_idx = 0 ; x_idx < meshDim_X; x_idx++){
						for (int arm_idx = 1 ; arm_idx <= 12  ; arm_idx++) {
							mesh_data_client[12*(meshDim_Y*meshDim_X*z_idx + meshDim_X*y_idx + x_idx)+  0]  = mesh_host[address(arm_idx,x_idx,y_idx,z_idx)];
						}
					}
				}
			}
		} else {
			printf(" GPU mesh structure needs to be created first\n");
		}
			*gpu_mesh_X = meshDim_X;
			*gpu_mesh_Y = meshDim_Y;
			*gpu_mesh_Z = meshDim_Z;
	}
	// Get GPU_Mesh Dimensions
	void GPU::Mesh_Dimensions_get(int *x, int *y, int *z){
			*x = meshDim_X;
			*y = meshDim_Y;
			*z = meshDim_Z;
	}
	// Get a Node from GPU_Mesh 
	// if armID ==0 then average of all 12 voltage arms in node are averaged then retrieved. 
	// if armID !=0 and between 1 and 12 then value of the arm in defined node is retrieved
	float GPU::Mesh_Node_get(int x, int y, int z, int armID ){
		float result = 0;
		
		
		if ((x>meshDim_X)|| (y>meshDim_Y)|| (z>meshDim_Z) ) {
			return 0;
		}
		if (armID <0 || armID> 12){
			return 0;
		}

		if (armID == 0) {			
			float tmp =0;
			for (int arm_idx = 1 ; arm_idx <=12; arm_idx++){
				tmp += mesh_host[address(arm_idx,x,y,z)];
			}
			result = tmp/12.0;
		} else {
			result = mesh_host[address(armID,x,y,z)];
		}

		return result;
	}


	/// SETs
	// Set a node with passed value 
	// if armID == 0 then all 12 voltage arms of node coord are set to node_value
	// if armID != 0 then value of arm in difined node is set 
	void GPU::Mesh_Node_set(float node_value, int x, int y, int z, int armID){
		
		if (armID == 0){
			for (int arm_idx = 1 ; arm_idx <=12; arm_idx++){
				mesh_host[address(arm_idx,x,y,z)] = node_value;
			} // end of for next
		} else {
			mesh_host[address(armID,x,y,z)] = node_value;
		} 
	
	}// End of method

	// Pass a plane of nodes in an array of data, "mesh_data" and set a plane in the GPU_Mesh. 
	// if armID ==0 then all 12 volage arms in each node falling on the defined plane, on GPU, will be set to CPU plane values
	// if armID != 0 and between 1 and 12 then value of the arm in dfined nodes of plane are set
	// dimensions must define a plane and not a 3D volume. 
	void GPU::Mesh_Node_plane_set(float* mesh_host, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi, int armID){

	}
	// Set a 3D volume in GPU_Mesh with values passed into method. (Note: all 12 arms of nodes in defined volume will be set
	void GPU::Mesh_Node_cube_set(float *mesh_host, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){
	}


	int GPU::CPU_to_GPU_address(int x, int y, int z){
		
		int grid_cubes_X = int(meshDim_X/BLOCK_SIZE_X); // number of cubes accross x dir
		if( (meshDim_X % BLOCK_SIZE_X) >0 ) grid_cubes_X++; // add one more cube if needed

		int grid_cubes_Y = int(meshDim_Y/BLOCK_SIZE_Y);
		if( (meshDim_Y % BLOCK_SIZE_Y) >0 ) grid_cubes_Y++;

		int grid_cubes_Z = int(meshDim_Z/BLOCK_SIZE_Z);
		if( (meshDim_Z % BLOCK_SIZE_Z) >0 ) grid_cubes_Z++;

		//  the following locate the 'cube' address in which x,y,z reside
		int cube_x = int(x/BLOCK_SIZE_X);  // which cube co-od (in x-direction) is x located 
		int cube_y = int(y/BLOCK_SIZE_Y);  // which cube co-od (in y-direction) is y located 
		int cube_z = int(z/BLOCK_SIZE_Z);  // which cube co-od (in z-direction) is z located 

		// Starting_address is the location (node address) of the first node of a 'cube' where x,y,z reside
		int starting_address = BLOCK_SIZE_Z*BLOCK_SIZE_Y*BLOCK_SIZE_X*cube_x;// x address (cube)
		starting_address += BLOCK_SIZE_Z*BLOCK_SIZE_Y*BLOCK_SIZE_X*grid_cubes_X*cube_y; // y address (cube row)
		starting_address += BLOCK_SIZE_Z*BLOCK_SIZE_Y*BLOCK_SIZE_X*grid_cubes_Y*grid_cubes_X*cube_z; // z address (slice of cubes)

		// the 'internal' cube address (node address) of the cube that x,y,z resides in. 
		//   (i.e. if a cube is 5x5x5 = 125 nodes/cube then:  0 < cube_inner_address < 124  ) 
		int cube_inner_address = x % BLOCK_SIZE_X;
		cube_inner_address += BLOCK_SIZE_Y* (y % BLOCK_SIZE_Y);
		cube_inner_address += BLOCK_SIZE_Z*BLOCK_SIZE_Y*(z%BLOCK_SIZE_Z);

	// i.e. if the 'starting_address' of the cube which x,y,z reside in is 2000, then the full address of x,y,z will be 
	//     between address 2000 and 20124 ( given a 5x5x5 =125 node cube). 
	return starting_address + cube_inner_address; 
	}


	void GPU::Mesh_Device_to_host_memcopy(){
		cuda_Device_to_host_memcopy_relay(GPU::GPU_Device_Mesh, GPU::mesh_host, (GPU::meshDim_X*	GPU::meshDim_Y*GPU::meshDim_Z));
	
	}

	// Sets constants in the GPU to be used for all mesh realted calculations (iteration, boundary, excitation...)
	// sets the following:  meshDim_X,meshDim_Y,meshDim_Z,GPU_BLOCK_X,GPU_BLOCK_Y,GPU_BLOCK_Z 
	void GPU::cuda_Mesh_Block_constants_set(){
		cuda_Mesh_Block_constants_set_relay(meshDim_X,meshDim_Y,meshDim_Z,GPU::GPU_BLOCK_X,GPU::GPU_BLOCK_Y,GPU::GPU_BLOCK_Z );
	}

	//  end of MESH methods
	/////////////////


	/////////////////
	// BOUNDARY METHODS
	//
	// add a boundary to Mesh structure
	void GPU::bndry_add(float ref_coef, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){
	}
	// get a boundary from Mesh structure by ID. (boundary must exist for this to return a boundary)
	void GPU::bndry_get(int bdry_index, float *ref_coef, int *xlow, int *xhi, int *ylow, int *yhi, int *zlow, int *zhi){
	}
	// Write over an existing boundary. (boundary ID must exist)
	void GPU::bndry_set(int bdry_index, float ref_coef, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){
	}
	//void GPU::bndry_seed_boundaries(){ // OBSOLETE
	//	int this_bndry_temp = 0;    
	//	int other_bndry_temp = 0;
	//	for (int z_idx = 0 ; z_idx < meshDim_Z; z_idx++){
	//		for (int y_idx = 0 ; y_idx < meshDim_Y; y_idx++){
	//			for (int x_idx = 0 ; x_idx < meshDim_X; x_idx++){
	//				// loop through all boundaries
	//				for (int bndry_idx = 0; bndry_idx < boundaries_Num; bndry_idx++){
	//					// Is this node within one of the boundaries (Inclusive)
	//					if(   (x_idx >= boundaries[bndry_idx].Xmin )
	//						&&(x_idx <= boundaries[bndry_idx].Xmax )
	//						&&(y_idx >= boundaries[bndry_idx].Ymin )
	//						&&(y_idx <= boundaries[bndry_idx].Ymax )
	//						&&(z_idx >= boundaries[bndry_idx].Zmin )
	//						&&(z_idx <= boundaries[bndry_idx].Zmax )
	//						){
	//							// test if on edge
	//							this_bndry_temp = 0;
	//							other_bndry_temp = 0;
	//							// The following tests if the structure is a flat surface and not a volume
	//							bool flat_surface_X_true = (boundaries[bndry_idx].Xmin == boundaries[bndry_idx].Xmax );
	//							bool flat_surface_Y_true = (boundaries[bndry_idx].Ymin == boundaries[bndry_idx].Ymax );
	//							bool flat_surface_Z_true = (boundaries[bndry_idx].Zmin == boundaries[bndry_idx].Zmax );
	//								
	//							
	//							// Embed if on X edge of surface. (But not if Xmin is the on a surface in the Y or Z direction. i.e. edge-on to a surface)
	//							if ( (x_idx == boundaries[bndry_idx].Xmin)&& (!flat_surface_Y_true)&& (!flat_surface_Z_true)){
	//								this_bndry_temp = 0x02;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx ,this_bndry_temp,    boundaries[bndry_idx].ref_coef_index);
	//								
	//								other_bndry_temp = 0x08;
	//								bndry_Embed_Boundary_Data(x_idx-1, y_idx, z_idx ,other_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//							}
	//							if ((x_idx == boundaries[bndry_idx].Xmax) && (boundaries[bndry_idx].Xmin != boundaries[bndry_idx].Xmax )  ){
	//								this_bndry_temp = 0x08;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx ,this_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//								
	//								other_bndry_temp = 0x02;
	//								bndry_Embed_Boundary_Data(x_idx+1, y_idx, z_idx ,other_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//							}



	//							if ((y_idx == boundaries[bndry_idx].Ymin)&& (!flat_surface_X_true)&& (!flat_surface_Z_true)){
	//								this_bndry_temp = 0x20;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx ,this_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//								other_bndry_temp = 0x10;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx-1, z_idx ,other_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//							}
	//							if ((y_idx == boundaries[bndry_idx].Ymax) && (boundaries[bndry_idx].Ymin != boundaries[bndry_idx].Ymax )  ){
	//								this_bndry_temp = 0x10;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx ,this_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//								other_bndry_temp = 0x20;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx+1, z_idx ,other_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//							}
	//							
	//							
	//							if ((z_idx == boundaries[bndry_idx].Zmin)&& (!flat_surface_X_true)&& (!flat_surface_Y_true)){
	//								this_bndry_temp = 0x01;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx ,this_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//								other_bndry_temp = 0x04;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx-1 ,other_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//							}
	//							
	//							if ((z_idx == boundaries[bndry_idx].Zmax) &&(boundaries[bndry_idx].Zmin != boundaries[bndry_idx].Zmax) ){
	//								this_bndry_temp = 0x04;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx ,this_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//								other_bndry_temp = 0x01;
	//								bndry_Embed_Boundary_Data(x_idx, y_idx, z_idx+1,other_bndry_temp, boundaries[bndry_idx].ref_coef_index);
	//							}

	//					} // end of if (within boundary)
	//				} // end of for/next (looping boundaries)


	//			} // end of for/next 
	//		}// end of for/next
	//	}// end of for/next
	//}

	//void GPU::bndry_Embed_Boundary_Data(int x, int y , int z,int bndry, int ref_coef_index){ // OBSOLETE
	//	int dir_bits;
	//	int ref_idx_1;
	//	int ref_idx_2;
	//	int ref_idx_3;
	//	int ref_idx_4;
	//	
	//	int insertion_pos =0;
	//	///////////////////////////////
	//	//  Embedd .ref_coef_index with direction (bndry) and index (bndry_code_idx)
	//	//
	//	// Get copy of existing code (but do not overwrite)
	//	int direction_code = 0;
	//	
	//	direction_code = (int)mesh_host[CPU_to_GPU_address(x,y,z)].Bndry_Dir;// & 0x3F;
	//	// Count the number of boundaries already embedded to find the position to write the 4 bit index
	//	int bndry_code_idx = 0;


	//	if (bndry&0x01) insertion_pos =1;
	//	if (bndry&0x02) insertion_pos =2;
	//	if (bndry&0x04) insertion_pos =3;
	//	if (bndry&0x08) insertion_pos =4;
	//	if (bndry&0x10) insertion_pos =5;
	//	if (bndry&0x20) insertion_pos =6;

	//	if ( (direction_code & 0x1) && (insertion_pos > 1 ) ){	// down
	//		bndry_code_idx++;
	//	}
	//	if ( (direction_code & 0x2) && (insertion_pos > 2 ) ) {   // left
	//		bndry_code_idx++;}
	//	if ( (direction_code & 0x4) && (insertion_pos > 3 ) )	{   // up 
	//		bndry_code_idx++;}
	//	if ( (direction_code & 0x8) && (insertion_pos > 4 ) ) {   // right
	//		bndry_code_idx++;}
	//	if ( (direction_code & 0x10) && (insertion_pos > 5 ) ) {  //forward
	//		bndry_code_idx++;}
	//	if ( (direction_code & 0x20) && (insertion_pos > 6 ) ) {  // back
	//		bndry_code_idx++;}
	//	if (bndry_code_idx >4) {
	//		printf(" Error!!!!:  Number of boundaries for node (%d,%d,%d) exceeds four! \n\n",x,y,z);
	//	}

	//	// Embed the direction bit 
	//	int temp_parse = 0;


	//	direction_code = direction_code | bndry; // first, splice in the direction bit 
	//	// Carve out each mapped variable (Direction Bits (6 bits), and 4 boundary indexes (4 bits ea.)
	//	dir_bits = (direction_code&   0x3F);// just carve out the direction bits and store them
	//	ref_idx_1 = (direction_code&    0x3C0)>>6;
	//	ref_idx_2 = (direction_code&   0x3C00)>>(6+4);
	//	ref_idx_3 = (direction_code&  0x3C000)>>(6+2*4);
	//	ref_idx_4 = (direction_code& 0x3C0000)>>(6+3*4);



	//	// no other direction bits present before the target insertion bit therefore 
	//	//   move positions 1,2,3 to 2,3,4 then splice in boundary index into pos 1.
	//	if (bndry_code_idx == 0){ 
	//		
	//		ref_idx_4 = ref_idx_3;
	//		ref_idx_3 = ref_idx_2;
	//		ref_idx_2 = ref_idx_1;
	//		ref_idx_1 = ref_coef_index; // 

	//		temp_parse = 0;
	//		temp_parse = temp_parse | dir_bits;
	//		temp_parse = temp_parse | (ref_idx_1<<(6));
	//		temp_parse = temp_parse | (ref_idx_2<<(6+4));
	//		temp_parse = temp_parse | (ref_idx_3<<(6+2*4));
	//		temp_parse = temp_parse | (ref_idx_4<<(6+3*4));

	//		direction_code =temp_parse;
	//		
	//	}
	//	
	//	if (bndry_code_idx == 1){ // move ref_coef_index positions 2,3 to 3,4 then insert ref_coef_index into position 2. Don't touch pos 1
	//		//printf("x=%d,  y=%d  , z=%d \n",x,y,z);
	//		
	//		ref_idx_4 = ref_idx_3;
	//		ref_idx_3 = ref_idx_2;
	//		ref_idx_2 = ref_coef_index;

	//		temp_parse = 0;
	//		temp_parse = temp_parse | dir_bits;
	//		temp_parse = temp_parse | (ref_idx_1<<(6));
	//		temp_parse = temp_parse | (ref_idx_2<<(6+4));
	//		temp_parse = temp_parse | (ref_idx_3<<(6+2*4));
	//		temp_parse = temp_parse | (ref_idx_4<<(6+3*4));

	//		direction_code =temp_parse;


	//		/*
	//		temp_parse = direction_code & 0x3FC00;  // Parse out boundary positions 2,3
	//		temp_parse = temp_parse<<4; // move to pos 3,4
	//		temp_parse = temp_parse | (ref_coef_index << (6+ 4*1));// splice in the new ref index into pos 2
	//		direction_code = direction_code & 0x3FF; //Blank out position 2,3,4
	//		direction_code = direction_code | temp_parse; // final splice of ref indexes 2,3,4
	//		*/
	//	}

	//	if (bndry_code_idx == 2){ // move ref_coef_index positions 3 to 4 then insert ref_coef_index into position 3. Don't touch pos 1,2
	//		
	//		temp_parse = direction_code & 0x3C000;  // Parse out boundary position 3
	//		temp_parse = temp_parse<<4; // move to pos 4
	//		temp_parse = temp_parse | (ref_coef_index << (6+ 4*2));// splice in the new ref index into pos 3
	//		direction_code = direction_code & 0x3FFF; //Blank out position 3,4
	//		direction_code = direction_code | temp_parse; // final splice of ref indexes 3,4
	//		
	//	}

	//	if (bndry_code_idx == 3){ // if three other dir bits exist therefore only one more can fit. So just splice ref_coef_index into last (4th) position
	//		direction_code = direction_code | (ref_coef_index << (6 +4*3));
	//	}

	//	// write final result back to node of interest
	//	mesh_host[CPU_to_GPU_address(x,y,z)].Bndry_Dir = (float)direction_code ;
	//	//
	//	////////////////////////
	//}

	

	void GPU::read_Structure_Definition_file(char* filename){
		float x, y;
		float t_min,t_max,delta_t=0;
		int num_t,idx = 0;
		float t=0;

		char ch, s[80];
		int n;
		FILE  *fi;  
		//fi = fopen("Data/STRUCUTRE_DEFINITIONS.csv", "rt");  /* note r as in read */
		fi = fopen(filename, "rt");  
		if(fi == NULL)
		{
			printf("Unable to open input file!\n");
			fcloseall();  /* close any open files */
			exit(-1);
		}
		int num_of_bndries=0;
		
		char * pch;

		fgets(s, 80, fi);  // get Mesh Dimensions
		pch = strtok (s,","); // throw away title 
		pch = strtok (NULL,",");
		meshDim_X = atof(pch);  // get mesh X dim
		pch = strtok (NULL,",");
		meshDim_Y = atof(pch); // get mesh Y dim
		pch = strtok (NULL,",");
		meshDim_Z = atof(pch); // get mesh Z dim
		
		fgets(s, 80, fi); // Resolution
		pch = strtok (s,","); // throw away title
		pch = strtok (NULL,","); // parse out resolution
		GPU::resolution = atof(pch);// assign resolution

		fgets(s, 80, fi); // boundary # 
		pch = strtok (s,",");
		pch = strtok (NULL,","); // get # of Boundaries 
		boundaries_Num = atoi(pch);// get # of Boundaries 
		boundaries = (BoundaryType*)malloc(boundaries_Num*sizeof(BoundaryType)); // allocate mem for boundaries

		fgets(s, 80, fi);  // Boundary Index Title
		  
		fgets(s, 80, fi);  // Boundary Values for indexes
		pch = strtok (s,","); // throw away title 
		//float reflection_coef_tmp[16];
		for (int idx = 0; idx < 16 ; idx++){ // read in Reflection coef (up to 16 max)
			pch = strtok (NULL,",");
			if (pch != NULL ) { 
				bndry_reflection_coef_list[idx] = atof(pch);
			}else{
				bndry_reflection_coef_list[idx] = 0;
			}
		}
		  
		// Get Boundary Dimensions
		fgets(s, 80, fi);  	// Throw away Title to boundary Definitions
		for (int idx = 0 ; idx < boundaries_Num; idx++){
			fgets(s, 80, fi);  	
			pch = strtok (s,","); // throw away title of boundary
			boundaries[idx].ref_coef_index = atoi(strtok (NULL,",")); 
			boundaries[idx].Xmin = atoi(strtok (NULL,",")); 
			boundaries[idx].Xmax = atoi(strtok (NULL,",")); 
			boundaries[idx].Ymin = atoi(strtok (NULL,",")); 
			boundaries[idx].Ymax = atoi(strtok (NULL,",")); 
			boundaries[idx].Zmin = atoi(strtok (NULL,",")); 
			boundaries[idx].Zmax = atoi(strtok (NULL,",")); 
			boundaries[idx].reflection_coef = bndry_reflection_coef_list[boundaries[idx].ref_coef_index ];
		}

		//sscanf(s, "#Boundaries =,%d,",&num_of_bndries);


	   fcloseall();
	} // end of read structure definitions


	void GPU::strucuture_scaling(float scale_strucutre){
		// Scaling 
		meshDim_X *= scale_strucutre;
		meshDim_Y *= scale_strucutre;
		meshDim_Z *= scale_strucutre;
		GPU::resolution = resolution/((float)scale_strucutre);
		for (int idx = 0 ; idx < boundaries_Num; idx++){
			boundaries[idx].Xmin *= scale_strucutre; 
			boundaries[idx].Xmax *= scale_strucutre; 
			boundaries[idx].Ymin *= scale_strucutre; 
			boundaries[idx].Ymax *= scale_strucutre; 
			boundaries[idx].Zmin *= scale_strucutre; 
			boundaries[idx].Zmax *= scale_strucutre; 
		}
	}


	void GPU::cuda_reflect_coef_constants_set(){
		cuda_reflect_coef_constants_set_relay(&bndry_reflection_coef_list[0], boundaries_Num);
	}
	void GPU::cuda_boundary_execute(int boundaryID){
		if ( boundaryID < GPU::boundaries_Num){
			cuda_boundary_execute_relay(GPU_Device_Mesh, boundaries[boundaryID].Xmin,boundaries[boundaryID].Xmax,boundaries[boundaryID].Ymin,boundaries[boundaryID].Ymax,boundaries[boundaryID].Zmin,boundaries[boundaryID].Zmax, boundaries[boundaryID].reflection_coef);
		}
	}
	// end of BOUNDRY Methods
	//////////////////


	/////////////////
	// ITERATION METHODS
	//
	void GPU::cuda_iteration_setup(){
		
		iterationSetup.meshDim_X = meshDim_X;
		iterationSetup.meshDim_Y = meshDim_Y;
		iterationSetup.meshDim_Z = meshDim_Z;

		iterationSetup.thread_BLK_X = BLOCK_SIZE_X; // 64
		iterationSetup.thread_BLK_Y = BLOCK_SIZE_Y; //  1
		iterationSetup.thread_BLK_Z = BLOCK_SIZE_Z; //  1

		iterationSetup.blocksPerXGrid = meshDim_X/(BLOCK_SIZE_X);
		if (meshDim_X%BLOCK_SIZE_X != 0) iterationSetup.blocksPerXGrid++;
		iterationSetup.blocksPerYGrid = (meshDim_Y*meshDim_Z);
		if ((iterationSetup.blocksPerYGrid>65535) ){ printf(" ERROR!!! blocksPerGrid_Y=%d This exceeds 65535. \n\n",iterationSetup.blocksPerYGrid); }
		
		//iterationSetup.nodes_per_Zslice = BLOCK_SIZE_Z*meshDim_X*meshDim_Y; // OBSOLETE
		//iterationSetup.slices_Per_Zdim = meshDim_Z/(BLOCK_SIZE_Z) ; // OBSOLETE
	    
		iterationSetup.threads = dim3( BLOCK_SIZE_X, BLOCK_SIZE_Y,BLOCK_SIZE_Z); 
		iterationSetup.grid = dim3(iterationSetup.blocksPerXGrid, iterationSetup.blocksPerYGrid);

	}

	// execute one or more iterations of simulation on the GPU
	void GPU::iteratate(){
		display_percent_done();
		cuda_iteration_relay(GPU::iteration_index, GPU::GPU_Device_Mesh,  &iterationSetup);
		GPU::iteration_index++;

	}
		
	void GPU::interchange_3D_mesh(){
		cuda_TLM_3D_InterchangeOnly(GPU::iteration_index, GPU::GPU_Device_Mesh,  &iterationSetup);
	}

	//  end of Iteration methods
	//////////////////
	
	//////////////////
	// EXCITATION 
	
	// Returns excitation Object ID
	int GPU::excitation_CreateStructure(int prop_dir, int polz_dir, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){	
		// check if polarization and propogation and structure align properly
		if( (prop_dir == 1) && (xlow != xhi)) {printf("Error:'excitation_CreateStructure(...)' Propogation direction and structure do not match\n"); exit(0);}
		if( (prop_dir == 2) && (ylow != yhi)) {printf("Error:'excitation_CreateStructure(...)' Propogation direction and structure do not match\n"); exit(0);}
		if( (prop_dir == 3) && (zlow != zhi)) {printf("Error:'excitation_CreateStructure(...)' Propogation direction and structure do not match\n"); exit(0);}
		int tmp = 0;
		tmp = excitation::excitation_obj_count;
		excite_obj[tmp] = new GPU_Space::excitation(prop_dir, polz_dir, xlow, xhi, ylow, yhi, zlow, zhi);
		tmp = excitation::excitation_obj_count;
		return (excitation::excitation_obj_count-1);
	}
	void GPU::excitation_generate_SineGaussian_array(int excite_obj_ID,float freq, float magnitude, float width, float peak_position,  int excite_array_size){
		if (excite_obj_ID < excitation::excitation_obj_count){
			
			excite_obj[excite_obj_ID]->generate_excitation_SineGaussian_array(freq, GPU::resolution, magnitude, width, peak_position, excite_array_size);
		} else{
			printf(" Error: specified excitation object ID is out of bounds\n");
		}
	
	}


	void GPU::excitation_Save(int excite_obj_ID,char* filename){
		if (excite_obj_ID < excitation::excitation_obj_count){
				excite_obj[excite_obj_ID]->save_excitation(filename);
		} else{
			printf(" Error: specified excitation object ID is out of bounds\n");
		}
		
	}



	void GPU::excitation_execute_Impulse(int x, int y, int z, float excitation){
		cuda_excitation_execute_Impulse_relay(GPU_Device_Mesh, x, y, z, excitation , GPU::iteration_index);
		//	cuda_excitation_execute_relay(GPU_Device_Mesh, excite_obj[excitation_ID]->scn_arm, excite_obj[excitation_ID]->xlow, excite_obj[excitation_ID]->xhi, excite_obj[excitation_ID]->ylow, excite_obj[excitation_ID]->yhi, excite_obj[excitation_ID]->zlow, excite_obj[excitation_ID]->zhi, 1,GPU::iteration_index);
	
	}
	
	void GPU::excitation_execute_HalfSine_Impulse(int excitation_ID){
		cuda_excitation_execute_relay(GPU_Device_Mesh, excite_obj[excitation_ID]->scn_arm, excite_obj[excitation_ID]->xlow, excite_obj[excitation_ID]->xhi, excite_obj[excitation_ID]->ylow, excite_obj[excitation_ID]->yhi, excite_obj[excitation_ID]->zlow, excite_obj[excitation_ID]->zhi, 1,GPU::iteration_index);
	}

	void GPU::excitation_execute_HalfSine(int excitation_ID){
		if (GPU::iteration_index < excite_obj[excitation_ID]->excite_array_size){
			cuda_excitation_execute_relay(GPU_Device_Mesh, excite_obj[excitation_ID]->scn_arm, excite_obj[excitation_ID]->xlow, excite_obj[excitation_ID]->xhi, excite_obj[excitation_ID]->ylow, excite_obj[excitation_ID]->yhi, excite_obj[excitation_ID]->zlow, excite_obj[excitation_ID]->zhi,excite_obj[excitation_ID]->excitation_data[iteration_index],GPU::iteration_index);
		}
	}
	

	//
	/////////////////

	///////////////
	// SAMPLING PROBES
	//

	// add a sampling probe object to GPU object 
	int GPU::sample_Probe_add(int sample_maxnum, int polerize_dir, int x, int y, int z){
		//int sampling_maxNum; 
		//sampling_probe* sampling_obj[100];
		int tmp = 0; 
		tmp = sampling_probe::sampling_obj_count;
		
		sampling_obj[tmp] = new GPU_Space::sampling_probe(sample_maxnum,polerize_dir,x,y,z);
		sampling_obj[tmp]->samples = new float[sample_maxnum];
		//sampling_obj[tmp]->sample_numMax = sample_maxnum;
		for(int idx = 0 ; idx < sample_maxnum ; idx++)sampling_obj[tmp]->samples[idx] = 0;
		sampling_obj[tmp]->sample_index = 0;
		sample_gpu_sampleArray_Allocate(tmp);
		tmp = sampling_probe::sampling_obj_count;
		return (sampling_probe::sampling_obj_count-1);
	}
	// retri
	void GPU::sample_Probe_get(int probe_index, float* data, int* sample_maxnumint, int armID){
	}


	void GPU::sample_execute(int sample_ID){
		if (sampling_obj[sample_ID]->sample_index < sampling_obj[sample_ID]->sample_numMax){
			cuda_sampling_execute_relay(GPU_Device_Mesh,sampling_obj[sample_ID]->gpu_samples , sampling_obj[sample_ID]->sample_index , sampling_obj[sample_ID]->sample_polarize_dir, sampling_obj[sample_ID]->x, sampling_obj[sample_ID]->y,sampling_obj[sample_ID]->z, GPU::iteration_index); 
		}
		sampling_obj[sample_ID]->sample_index++;
	}


	void GPU::sample_gpu_sampleArray_Allocate(int sample_ID){
		cuda_gpu_sampleArray_Allocate_relay(&GPU::sampling_obj[sample_ID]->gpu_samples, sampling_obj[sample_ID]->sample_numMax);
	}

	// get samples from GPU ; if num_of_samples is out of bounds then "sample_numMax" are retrieved 
	float* GPU::sample_gpu_samples_get(int sample_ID,int num_of_samples){
		if( (num_of_samples <= 0) || (num_of_samples > sampling_obj[sample_ID]->sample_numMax) ) num_of_samples = sampling_obj[sample_ID]->sample_numMax;
		cuda_Samples_get_relay(sampling_obj[sample_ID]->gpu_samples,  sampling_obj[sample_ID]->samples, num_of_samples);
		return sampling_obj[sample_ID]->samples;
	}

	void GPU::sample_save_to_file(char* filename, int sample_ID){
		cuda_save_results_to_CSVfile(filename, sampling_obj[sample_ID]->samples, sampling_obj[sample_ID]->sample_numMax, GPU::delta_time);
	}

	// end of SAMPLING Probes
	///////////////

	/////////////////////////////
	// Misc Methods
	//void GPU::display_percent_done(int iter, int data_cpu_size_inFloats, int numofiterations, int total_Nodes, timer* timer_ptr){
	void GPU::display_percent_done(){
		if (iteration_timer_display_ENABLE == true){
			if ( ( iteration_timer.elapsed_time() - iteration_timer_lastTime ) > 1){
				double elapsedTime = iteration_timer.elapsed_time();
				double MegaNodes_per_sec_overall = ((double)(meshDim_X*meshDim_Y*meshDim_Z)*(double)GPU::iteration_index) / ((double)elapsedTime*(double)1000000) ;
				double MegaNodes_per_sec_instant = ((double)(meshDim_X*meshDim_Y*meshDim_Z)*(double)(GPU::iteration_index-iteration_previously_saved_idx)) / ((double)(elapsedTime-iteration_timer_lastTime)*(double)1000000) ;
				double MegaNodes_per_sec_running_avg  =  0.999*MegaNodes_per_sec_instant + 0.001*MegaNodes_per_sec_running_avg_savd ;

				double gbytes_per_sec =  ( (double)(4*12*meshDim_X*meshDim_Y*meshDim_Z)*(double)GPU::iteration_index) / ((double)elapsedTime*(double)1000000000) ;
				
				if (MegaNodes_per_sec_instant>1e3) MegaNodes_per_sec_instant = 1e3;
				if (MegaNodes_per_sec_running_avg>1e3) MegaNodes_per_sec_running_avg = 1e3;
				MegaNodes_per_sec_running_avg_savd = MegaNodes_per_sec_running_avg;
				iteration_timer_lastTime = elapsedTime;
				iteration_previously_saved_idx = GPU::iteration_index;
				//printf("%%done: ,%4.1f,%% iteration:,%d, Rate_OverallAvg: ,%4.1f, or Instantaneous: ,%4.1f,(MNodes/s)\n",100*((double)iter)/((double)iterations_MaxNum),iter,MegaNodes_per_sec_overall,MegaNodes_per_sec_running_avg);
				//printf(" Nodes: %d, %%done: ,%4.1f,%%, iteration:,%d, Rate: ,%4.1f (MNodes/s), %4.3f(GBytes/sec) \r ",meshDim_X*meshDim_Y*meshDim_Z ,100*((double)GPU::iteration_index)/((double)iterations_MaxNum),GPU::iteration_index,MegaNodes_per_sec_overall,gbytes_per_sec);
				printf("  %%done: ,%4.1f,%%, iteration:,%d, Rate: ,%4.1f (MNodes/s), %4.3f(GBytes/sec) \r ",100*((double)GPU::iteration_index)/((double)iterations_MaxNum),GPU::iteration_index,MegaNodes_per_sec_overall,gbytes_per_sec);
			}
		}
	} // end of method


	void GPU::Output_Node_Layer_XY_Plane(char* filename, int z_layer, int voltage_arm ){
	//~~~
		FILE *OutFile = fopen(filename,"w");
		for (int y_idx = 0 ; y_idx < GPU::meshDim_Y ; y_idx++) fprintf(OutFile,"%d,",y_idx);
		fprintf(OutFile,"\n");
		fprintf(OutFile,"Output of Average Node slice\n meshDim_X=,%d, meshDim_Y=,%d, Zlayer=,%d \n",GPU::meshDim_X, GPU::meshDim_Y  ,z_layer );
		fprintf(OutFile,"iteration = %d \n.\n,",GPU::iteration_index);
		fprintf(OutFile," Y------->\n");
		for (int y_idx = 0 ; y_idx < GPU::meshDim_Y ; y_idx++){
			fprintf(OutFile,"%d,",y_idx);
		}
		fprintf(OutFile,".\n");
		float average =0;
		for (int x_idx = 0; x_idx < GPU::meshDim_X ; x_idx++){
			for (int y_idx = 0 ; y_idx < GPU::meshDim_Y ; y_idx++){
				if(voltage_arm == 0 ){
					for (int armID = 1 ; armID <= 12 ; armID++){
						average += GPU::mesh_host[address(armID,x_idx,y_idx,z_layer)];
					} // end of armID for next
					average = average/12.0;
					fprintf(OutFile,"%f,",average);
					average = 0;
				} else {
					fprintf(OutFile,"%f,",GPU::mesh_host[address(voltage_arm,x_idx,y_idx,z_layer)]);
				} // end of if
			}// end of y_idx for next
			fprintf(OutFile,"\n");
		} // end of x_idx for next

		//Close the file 
		fclose(OutFile);
	} // end of method

 
	// end of Misc Methods
	////////////////










	//////////////////
	// Private Methods
	//
	// In-place format conversion from CPU mesh type to GPU Mesh type
	//   Extra memory required for swapping not known as this time
	void GPU::CPUmesh_to_GPUmesh_convert(){
	}
	// In-place format conversion from GPU mesh type to CPU Mesh type
	//   Extra memory required for swapping not known as this time
	void GPU::GPUmesh_to_CPUmesh_convert(){
	}
	// insert all boundaries definitions into mesh (Mesh type includes a float for boundaries)
	void GPU::bndry_seed(int bndry_maxNum, GPU_Space::GPU::BoundaryType *boundaries){
	}
	
	
	
	// end of Private Methods
	//////////////////


	/////////////////////////////////////
	// CLASS sampling_probe Definitions//
	/////////////////////////////////////
	//Constructor
	int sampling_probe::sampling_obj_count = 0;
	sampling_probe::sampling_probe(){
	}

	sampling_probe::sampling_probe(int sample_maxnum, int polerize_dir, int x, int y, int z){
		sampling_probe::sampling_obj_count++;
		sampling_probe::x = x;
		sampling_probe::y = y;
		sampling_probe::z = z;
		sampling_probe::sample_polarize_dir = polerize_dir;
		sampling_probe::sample_numMax = sample_maxnum;
	}
	



//////////////////////////////////////////////////////

	/////////////////////////////////
	// Class Excitation Definitions//
	/////////////////////////////////
	// Excitation Object is used to set excitation values in the GPU Mesh (residing on the GPU device) 
	// The excitation parameters can be generated or read from file (or saved to file). 
	// The excation dimensional profile defines the type and location of the exciation point/plane/cube
	// the excitation time profile defines the frequency and wavform of excitation
	//Constructors
	int excitation::excitation_obj_count = 0;
	excitation::excitation(){
		excitation_obj_count++;
		
	}

	// explicitely defines the demension of the excitation plane in the mesh
	excitation::excitation(int prop_dir, int polz_dir, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){
		excitation::ojb_ID = excitation_obj_count;
		excitation_obj_count++;
		// Find scn_arm
			//ALL = 0, XposYpol = 11, XnegYpol = 3, XposZpol = 10, XnegZpol = 6, 
	//		   YposXpol = 12, YnegXpol = 1, YposZpol =  7, YnegZpol = 5, 
	//		   ZposXpol =  9, ZnegXpol = 2, ZposYpol =  8, ZnegYpol = 4

	//propogation_dir: // NONE = 0, X+ve =  1, Y+ve =  2, Z+ve =  3
		//							X-ve = -1, Y-ve = -2, Z-ve = -3
	//polorize_dir: //NONE =0, X = 1, Y = 2, Z = 3
		excitation::propogation_dir = prop_dir;
		excitation::polorize_dir = polz_dir;
		
		if (polz_dir == 1){
			if ( prop_dir ==  -2 ) scn_arm = 12 ;  //YnegXpol = 12
			else if ( prop_dir == 2 ) scn_arm = 1  ;  //YposXpol =  1
			else if ( prop_dir ==  -3 ) scn_arm = 9  ;  //ZnegXpol =  9
			else if ( prop_dir == 3 ) scn_arm = 2  ;  //ZposXpol =  2
		} else if (polz_dir == 2){
			if ( prop_dir ==   -1 ) scn_arm = 11 ;  //XnegYpol = 11
			else if ( prop_dir ==  1 ) scn_arm =  3 ;  //XposYpol =  3
			else if ( prop_dir ==   -3 ) scn_arm =  8 ;  //ZnegYpol =  8
			else if ( prop_dir ==  3 ) scn_arm =  4 ;  //ZposYpol =  4
		} else if (polz_dir == 3){
			if ( prop_dir ==   -1 ) scn_arm = 10 ;  //XnegZpol = 10
			else if ( prop_dir ==  1 ) scn_arm =  6 ;  //XposZpol =  6 
			else if ( prop_dir ==   -2 ) scn_arm =  7 ;  //YnegZpol =  7
			else if ( prop_dir ==  2 ) scn_arm =  5 ;  //YposZpol =  5 
		} else {
			scn_arm =0;
		}
		//if ( prop_dir ==  0 ) scn_arm = 0 ;  // No polerization

		//excitation::freq = freq;
		//excitation::dT  = dT; 
		//excitation::dNode = dNode;
		excitation::xlow = xlow;
		excitation::xhi  = xhi;
		excitation::ylow = ylow;
		excitation::yhi  = yhi;
		excitation::zlow = zlow;
		excitation::zhi  = zhi;

	}
	// destructor
	excitation::~excitation(){
		excitation_obj_count--;
	}
	
	


	// Dimension for point excitation 
	void excitation::point_set(int x, int y, int z){} // for point excitation
	// Dimension for line excitation (coordinates must form line and not point,plane or volume)
	void excitation::line_set(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){}

	// Dimension for plane excitation (coordinates must form plane and not point,line or volume)
	//ALL = 0, XposYpol = 11, XnegYpol = 3, XposZpol = 10, XnegZpol = 6, 
	//		   YposXpol = 12, YnegXpol = 1, YposZpol =  7, YnegZpol = 5, 
	//		   ZposXpol =  9, ZnegXpol = 2, ZposYpol =  8, ZnegYpol = 4
	void excitation::plane_set( int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){}
	
	// Dimension for volume excitation (coordinates must form volume and not point,line or plane)
	void excitation::cube_set(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi){}

	// Retrieve the point excitation coords (if applicable) 
	void excitation::point_get(int* x, int* y, int* z){}
	// Retrieve the line excitation coords  (if applicable) 
	void excitation::line_get( int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi){}
	// Retrieve the plane excitation coords  (if applicable) 
	void excitation::plane_get( int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi){}
	// Retrieve the volume excitation coords  (if applicable) 
	void excitation::cube_get(int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi){}
	
	// Set Frequency (GHz)
	void excitation::freq_set(float freq){}
	// Set delta time (sec)
	void excitation::delta_time_set(float dT){}
	// Set delta node dist (mm)
	void excitation::delta_node_dist_set(float dNode){}
	// Set number of excitation points (in time steps) Note: may be auto generated in future
	void excitation::excite_points_num_set(int excite_points_num){}

	// Get Freq (GHz)
	float excitation::freq_get(){return 0;}
	// Get delta time (sec)
	float excitation::delta_time_get(){return 0;}
	// Get delta node dist (mm)
	float excitation::delta_node_dist_get(){return 0;}
	// Get number of excitation points (in time steps) Note: may be auto generated in future
	int excitation::excite_points_num_set(){return 0;}
	
	////////////////
	// Private: 
	// Once all the parameters have been set this method will generate an exciation array read to be used by GPU
	void excitation::generate_excitation_SineGaussian_array(float freq, float dNode, float magnitude, float width, float peak_position, int excite_array_size_p){
		
		excitation::freq = freq;
		excite_array_size = excite_array_size_p;
		excitation::excitation_data	 = new float[excite_array_size]; // need to check if already instantiated

		//float VEL_LIGHT = 3E8; // m/sec
		double sampling_period = (dNode)/(VEL_LIGHT*2); // sec
		double sampling_freq = 1.0/sampling_period; // Hz
		double dFreq  = sampling_freq; //delta Hz
		
		excitation::dT = sampling_period;
		excitation::dFreq =sampling_freq; 
		double guassian,freq_coef;
	
		 // note: at delta_T = 1.09907e-012 and Target_freq = 31E9 Hz the FFT peak is at index 20. 31E9/20 = 1.55E+09 Hz/step

		freq_coef = 2*PI*freq*sampling_period;  //0.297; // Frequency coefficient to control the sinusoidal frequency  = 2*pi()*Freq*delta_T
		//phase_offset -> offsets the sinusoid in order to centre peak of sin() with peak of gaussian 

		for( int idx = 0 ; idx < excite_array_size ; idx ++ ) 
		{
			//phase_offset
			guassian = magnitude * exp( -pow(float(idx) - float(peak_position),2) / (2*pow(float(width),2) ));
			excitation_data[idx] = sin(freq_coef*float(idx) )*guassian;
		}
	}
	
	// read existing excitation profile from .csv file
	void excitation::excite_file_read(char fileName){
	}

	void excitation::save_excitation( char* filename){
	//cuda_save_results_to_CSVfile(char* filename, float samples_CUDA[], int sample_len, float delta_time){
	//FILE *OutFile = fopen("NormFFTData.csv","w");
	FILE *OutFile = fopen(filename,"w");
	
	if (!OutFile)
	{
		fprintf(stderr, "Error opening %s\nHit <Enter> to terminate the program\n", filename);
		fgetc(stdin);
		exit(0);
	}
	//fprintf(OutFile,"Grid=%dx%d , iterations=,%d  \n", V_X_SIZE,V_Y_SIZE,NUMOFITERATIONS  );
	fprintf(OutFile,"Excitation Index # %d \n", excitation::ojb_ID); //TBD pass index number
	
	char str_direction[20] = "---";
	if (excitation::propogation_dir == 0 )strcpy(str_direction,"No-Propagation");
	if (excitation::propogation_dir == 1 )strcpy(str_direction,"X-Propagation");
	if (excitation::propogation_dir == 2 )strcpy(str_direction,"Y-Propagation");
	if (excitation::propogation_dir == 3 )strcpy(str_direction,"Y-Propagation");

	char str_polarization[20] = "---";
	if (excitation::polorize_dir== 0 )strcpy(str_polarization,"No-Polarization");
	if (excitation::polorize_dir== 1 )strcpy(str_polarization,"X-Polarization");
	if (excitation::polorize_dir== 2 )strcpy(str_polarization,"Y-Polarization");
	if (excitation::polorize_dir== 3 )strcpy(str_polarization,"Y-Polarization");

	fprintf(OutFile,"Propogation Direction = %d  =  %s \n", excitation::propogation_dir,str_direction); 
	fprintf(OutFile,"Polarization Direction = %d =  %s \n", excitation::polorize_dir,str_polarization); 
	fprintf(OutFile," Xlow, Xhi, Ylow, Yhi, Zlow, Zhi \n");
	fprintf(OutFile," %d, %d,%d,%d,%d,%d \n\n",xlow, xhi, ylow, yhi, zlow, zhi);


	fprintf(OutFile,"Index, Time(sec) ,Excitation , \n");
	for (int i = 0; i < excitation::excite_array_size; i++) {
		//Send data to file
		//fprintf(OutFile,"%f , %f , %f \n",samples_CUDA[i],fft_results[i]); // write: Time,fft_mag, fft_phase
		float sample_time = ((float)i)*excitation::dT;
		fprintf(OutFile,"%d, %E , %E  \n",i, sample_time, excitation::excitation_data[i]); 
	}
	//Close the file 
	fclose(OutFile);
	}

		
	//
	//////////////////
	





} // end of namespace